from abc import abstractmethod

class CacheProvider():
    def __init__(self):
        raise NotImplementedError()

    @abstractmethod
    def set(self,key,data):
        pass

    @abstractmethod
    def get(self,key):
        pass



from .context import ApplicationContext
import inspect

class Autowired():

    def __init__(self,**kargs):
        self.config = kargs

    def __call__(self,beanMethod):
        def constructor(*args, **kargs):
            spec = inspect.getfullargspec(beanMethod)
            argsList = spec[0]
            annotations = spec[6]
            realargs=[args[0]]
            for arg in argsList:
                if arg == 'self' :
                    continue
                if self.config.get(arg) != None:
                    bean = ApplicationContext.resolve(self.config.get(arg),annotations.get(arg))
                else:
                    bean = ApplicationContext.resolve(arg,annotations.get(arg))
                realargs.append(bean)
            return beanMethod(*realargs,**kargs)
        return constructor



from .context import ApplicationContext
import inspect
class Bean():
    def __init__(self,**kargs):
        self.name = kargs.get("name","")
        self.initargs = kargs.get("args","")

    def __call__(self,classObj):
        name = self.name if self.name != "" else classObj.__name__
        classObj.__bean_name__ = name
        ApplicationContext.addBeanDefinition(classObj,self.initargs)
        return classObj



context={}

import os

for item in os.environ:
    if item.startswith("pspring"):
        os.environ[item.replace("_",".")] = os.environ.get(item)

from .core.bean import *
from .core.autowired import *
from .core.context import *
from .core.config import *
from .core.configprovider import *
from .core.cacheprovider import *



import inspect, types

from .config import Configuration

class Context():
    context = None
    def __init__(self):
        if Context.context == None:
            Context.context = {"byQualifier":{},"byType":{},"definitions":{}}

    def clear(self):
        Context.context.clear()
        Context.context = {"byQualifier":{},"byType":{},"definitions":{}}

    def addBeanDefinition(self,beanObj,args=None):
        Context.context.get("definitions").update({
            beanObj.__bean_name__ : (beanObj,args)
        })

    def getClassByName(self,name):
        return Context.context.get("byQualifier").get(name,None)

    def getClassByType(self,type):
        return Context.context.get("byType").get(type,None)

    def getClassThatDefinedMethod(meth):
        if inspect.ismethod(meth):
            for cls in inspect.getmro(meth.__self__.__class__):
               if cls.__dict__.get(meth.__name__) is meth:
                    return cls
            meth = meth.__func__  # fallback to __qualname__ parsing
        if inspect.isfunction(meth):
            cls = getattr(inspect.getmodule(meth),
                          meth.__qualname__.split('.<locals>', 1)[0].rsplit('.', 1)[0])
            if isinstance(cls, type):
                return cls
        return getattr(meth, '__objclass__', None) 

    def resolve(self,name,type):
        inst = self.getClassByName(name)
        if inst != None:
            return inst
        else:
            instArray = self.getClassByType(type)
            if instArray!=None and len(instArray) > 1:
                raise Exception("More than one instance found for type " +str(type))
            else:
                inst = instArray[0]
            if inst == None and Context.context.get("definitions").get(name,None) != None:
                return self.createBean(name,Context.context.get("definitions").get(name))
            else:
                return inst

    def registerByName(self,name,inst):
        Context.context.get("byQualifier").update({
            name : inst
        })

    def registerByType(self,inst):
        typeList = inspect.getmro(type(inst))
        for typeName in typeList:
            if typeName == type(object()):
                continue

            if Context.context.get("byType").get(typeName,None) == None:
                Context.context.get("byType")[typeName]=[]

            Context.context.get("byType")[typeName].append(inst)

    def createBeanFromMethod(self,beanClass,beanArgs):
        argspec = inspect.getfullargspec(beanClass)[0]
        args = [];
        for i in range(len(argspec)):
            #Skip the self args for methods bound to class
            #
            if(argspec[i]=="self"):
                args.append(self.getClassThatDefinedMethod(beanClass)())

            #Check if the args is provided
            #
            if beanArgs.get(argspec[i]) != None:
                args.append(beanArgs.get(argspec[i]))

            #Default args
            else:
                args.append(None)

        #Create the instance of the bean
        #
        inst = beanClass(*args)

        return inst

    def createBeanFromClass(self,beanClass,beanArgs):
        argspec = inspect.getfullargspec(beanClass.__init__)[0]
        args = [];
        for i in range(len(argspec)):
            #Skip the self args for methods bound to class
            #
            if(argspec[i]=="self"):
                continue

            #Check if the args is provided
            #
            if beanArgs.get(argspec[i]) != None:
                args.append(beanArgs.get(argspec[i]))

            #Default args
            else:
                args.append(None)

        #Create the instance of the bean
        #
        inst = beanClass(*args)

        return inst

    def createBean(self,beanName,beanObj):
        beanClass = beanObj[0]
        beanArgs = beanObj[1]
        inst = None
        if type(beanClass) == types.FunctionType:
            inst = self.createBeanFromMethod(beanClass,beanArgs)
        else:
            inst = self.createBeanFromClass(beanClass,beanArgs)
        
        if (inst != None):
            self.registerByName(beanName,inst)
            self.registerByType(inst)


    def initialize(self):
        definitions = Context.context.get("definitions")
        for beanName in definitions:
            self.createBean(beanName,definitions[beanName])

ApplicationContext = Context()



import inspect
import os
from .configprovider import ConfigurationProvider

class _ConfigInstance():
    def __init__(self,name,config):
        self.name = name
        self.config = config
        config.subscribe(self.callback)
        self.subscriptions = []

    def getProperty(self,propertyName):
        return self.config.getProperty(self.name+"."+propertyName)

    def callback(self):
        for subscription in self.subscriptions:
            subscription()

    def subscribe(self,callback):
        self.subscriptions.append(callback)

class Configuration():
    _defaults = {}
    _config = {}
    _configProviders = []
    _subscriptions = []

    def __init__(self):
        raise Exception("This is a singleton")

    @staticmethod
    def getConfig(moduleName):
        return _ConfigInstance(moduleName,Configuration)

    @staticmethod
    def subscribe(callback):
        Configuration._subscriptions.append(callback)

    @staticmethod
    def callback():
        for subscription in Configuration._subscriptions:
            subscription()

    @staticmethod
    def defaults(propertyMap):
        Configuration._defaults.update(propertyMap)

    @staticmethod
    def getProperty(propertyName):
        propertyScopeArray = propertyName.split(".")
        propertyScopes = []

        configProperty = propertyScopeArray.pop()
        propertyScopes.append(configProperty)
        fullPropertyScope=""

        for propertyScope in propertyScopeArray:
            fullPropertyScope = fullPropertyScope+propertyScope + "."
            propertyScopes.append(fullPropertyScope+configProperty)

        for derivedPropertyName in propertyScopes:
            if os.environ.get(derivedPropertyName) != None:
                return os.environ.get(derivedPropertyName)
            else:
                for bean in Configuration._configProviders:
                    propertyValue = bean.getProperty(derivedPropertyName)
                    if propertyValue != None:
                        return propertyValue

            if Configuration._defaults.get(derivedPropertyName) != None:
                return Configuration._defaults.get(derivedPropertyName)

        return None

    @staticmethod
    def clear():
        return Configuration._config.clear()

    @staticmethod
    def refresh():
        for bean in _configProviders:
            bean.refresh()

    @staticmethod
    def initialize(_configProviders):
        Configuration._configProviders = _configProviders
        #_configProviders = ApplicationContext.getClassByType(ConfigurationProvider)
        for bean in _configProviders:
            bean.subscribe(Configuration.callback)



from abc import abstractmethod

class ConfigurationProvider():
    def __init__(self):
        pass

    @abstractmethod
    def getProperty(self,propertyName):
        return None

    @abstractmethod
    def subscribe(self,callback):
        pass

    @abstractmethod
    def refresh(self):
        pass
